Package: exactextractr
Title: Fast Extraction from Raster Datasets using Polygons
Version: 0.10.1
Authors@R: c(
        person("Daniel", "Baston", email = "dbaston@isciences.com", role = c("aut", "cre")),
        person("ISciences, LLC", role="cph"))
Description: Quickly and accurately summarizes raster values over polygonal areas ("zonal statistics").
Depends: R (>= 3.4.0)
License: Apache License (== 2.0)
SystemRequirements: GEOS (>= 3.5.0)
Imports: Rcpp (>= 0.12.12), methods, raster, sf (>= 0.9.0),
URL: https://isciences.gitlab.io/exactextractr/,
        https://github.com/isciences/exactextractr
BugReports: https://github.com/isciences/exactextractr/issues
LinkingTo: Rcpp
Suggests: dplyr, foreign, knitr, ncdf4, rmarkdown, testthat, terra (>=
        1.5.17)
Encoding: UTF-8
RoxygenNote: 7.3.2
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2025-12-01 15:52:14 UTC; dan
Author: Daniel Baston [aut, cre],
  ISciences, LLC [cph]
Maintainer: Daniel Baston <dbaston@isciences.com>
Repository: CRAN
Date/Publication: 2025-12-01 16:50:18 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-27 22:13:10 UTC; unix
Archs: exactextractr.so.dSYM
