# Generated by roxygen2: do not edit by hand

exportMethods(coverage_fraction)
exportMethods(exact_extract)
exportMethods(exact_resample)
exportMethods(rasterize_polygons)
import(raster)
import(sf)
importFrom(Rcpp,evalCpp)
importFrom(methods,setMethod)
useDynLib(exactextractr)
