# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

CPP_coverage_fraction <- function(rast, wkb, crop) {
    .Call('_exactextractr_CPP_coverage_fraction', PACKAGE = 'exactextractr', rast, wkb, crop)
}

CPP_exact_extract <- function(rast, rast_ext, rast_res, rast_uncropped, weights, wkb, default_value, default_weight, include_xy, include_cell_number, include_area, area_weights, coverage_areas, p_area_method, include_cols, src_names, p_weights_names, warn_on_disaggregate, grid_compat_tol) {
    .Call('_exactextractr_CPP_exact_extract', PACKAGE = 'exactextractr', rast, rast_ext, rast_res, rast_uncropped, weights, wkb, default_value, default_weight, include_xy, include_cell_number, include_area, area_weights, coverage_areas, p_area_method, include_cols, src_names, p_weights_names, warn_on_disaggregate, grid_compat_tol)
}

CPP_stats <- function(rast, rast_ext, rast_res, weights, wkb, default_value, default_weight, coverage_areas, p_area_method, stats, max_cells_in_memory, grid_compat_tol, quantiles) {
    .Call('_exactextractr_CPP_stats', PACKAGE = 'exactextractr', rast, rast_ext, rast_res, weights, wkb, default_value, default_weight, coverage_areas, p_area_method, stats, max_cells_in_memory, grid_compat_tol, quantiles)
}

CPP_update_max_coverage <- function(extent, res, max_coverage, max_coverage_index, tot_coverage, wkb, index) {
    invisible(.Call('_exactextractr_CPP_update_max_coverage', PACKAGE = 'exactextractr', extent, res, max_coverage, max_coverage_index, tot_coverage, wkb, index))
}

CPP_resample <- function(rast_in, rast_out, p_stat, p_fun, coverage_area, area_method) {
    .Call('_exactextractr_CPP_resample', PACKAGE = 'exactextractr', rast_in, rast_out, p_stat, p_fun, coverage_area, area_method)
}

